/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */
#define __NO_VERSION_

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#include "wrapper.h"
#include "lib.h"
#include "sockio.h"
#include "pvfs_config.h"

#ifndef PVFS_NR_OPEN
#define PVFS_NR_OPEN (NR_OPEN)
#endif

extern fdesc_p pfds[];

/* IODINFO structure - this is used to keep up with all the IOD's in use
 *
 * DESCRIPTION:
 *
 * The idea here is to try to keep a single connection open to each
 * iod that we are communicating with.  To do this, we're going to
 * keep a little table and do our matching by comparing the port and 
 * address that we would use to connect to the manager.
 * We're going to vary dynamically the number of connections around.
 *
 */

/*
 * initial number of iod's for which we 
 * keep information about in this table
 */

enum {INITIAL_ENTRIES = 8, GROW_AMOUNT = 8};

static pid_t iodpid = 0;
/* Ahhhhh... the joys of single-threaded programming */
static int iodcnt = 0, initial_entries = 0, grow_amount = 0;
struct iodinfoitem {
	int      valid:1;
	unsigned short int port;
	unsigned int addr;
	int      fd;
	int 		ref_count;/* how many files are using this socket fd */
};
static struct iodinfoitem *iodinfo = NULL;


/* PROTOTYPES */
static int initiodconn(pid_t pid);

int init_iodtable(int _initial_entries, int _grow_amount)
{
	if(!iodinfo) {
		initial_entries = (_initial_entries <= 0) ? INITIAL_ENTRIES : _initial_entries;
		grow_amount 	 = (_grow_amount <= 0) ? GROW_AMOUNT : _grow_amount;
		iodinfo = (struct iodinfoitem *)kmalloc(initial_entries * sizeof(struct iodinfoitem), GFP_KERNEL);
		if(!iodinfo) {
			return -ENOMEM;
		}
		memset(iodinfo, 0, initial_entries * sizeof(struct iodinfoitem));
		/* none of the entries are valid initially */
		iodpid = current->pid;
		return 0;
	}
	return -EINVAL;
}

void cleanup_iodtable(void)
{
	int i;
	if(!iodinfo) {
		return;
	}
	for(i=0;i < initial_entries; i++) {
		if(iodinfo[i].fd > 0) {
			Close(iodinfo[i].fd);
		}
		iodinfo[i].fd = -1;
	}
	kfree(iodinfo);
	iodpid = 0;
	return;
}

static int grow_iodtable(void)
{
	size_t newsize;
	struct iodinfoitem *newiodinfo;
	newsize = (initial_entries + grow_amount) * sizeof(struct iodinfoitem);
	newiodinfo = (struct iodinfoitem *)kmalloc(newsize, GFP_KERNEL);
	if(!newiodinfo) {
		return -ENOMEM;
	}
	memset(newiodinfo, 0, newsize);
	memcpy(newiodinfo, iodinfo, initial_entries * sizeof(struct iodinfoitem));
	kfree(iodinfo);
	iodinfo = newiodinfo;
	initial_entries += grow_amount;
	return 0;
}

static int search_iodtable(struct sockaddr *saddr_p, int *freeslot)
{
	int i;
	if(freeslot) {
		*freeslot = initial_entries;
	}
	for(i=0; i < initial_entries; i++) {
		/* valid entry */
		if(iodinfo[i].valid != 0) {
			/* if address and port matches we return the sock fd handle */
			if(iodinfo[i].addr == ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr && iodinfo[i].port == ((struct sockaddr_in *)saddr_p)->sin_port) {
				return i;
			}
		}
		/* invalid entry ==> means free slot also */
		else {
			if(freeslot) {
				if(*freeslot > i) {
					*freeslot = i;		  
				}
			}
		}
	}
	return -1;
}

/* Just return the slot for a particular host */
int find_iodtable(struct sockaddr *saddr_p)
{
	int i;
	pid_t mypid;
	if(!iodinfo) {
		return -EINVAL;
	}
	if((mypid = current->pid) != iodpid) {
		/* detect fork */
		initiodconn(mypid);
		return -1;
	}
	for(i=0;i < initial_entries; i++) {
		if(iodinfo[i].valid != 0) {
			if(iodinfo[i].addr == ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr && iodinfo[i].port == ((struct sockaddr_in *)saddr_p)->sin_port) {
				return i;
			}
		}
	}
	return -1;
}

/* get the socket file descriptor for a given iod slot */
int getfd_iodtable(int slot)
{
	pid_t mypid;
	if(!iodinfo) {
		return -EINVAL;
	}
	/* invalid slot index */
	if(slot < 0 || slot >= initial_entries) {
		return -1;
	}
	if((mypid = current->pid) != iodpid) {
		/* detect fork */
		initiodconn(mypid);
		return -1;
	}
	/* valid slot */
	if(iodinfo[slot].valid != 0) {
		return iodinfo[slot].fd;
	}
	return -1;
}

/* given the socket file descriptor, what is the iod slot number */
int getslot_iodtable(int sockfd)
{
	int i;
	pid_t mypid;
	if(!iodinfo) {
		return -EINVAL;
	}
	if((mypid = current->pid) != iodpid) {
		/* detect fork */
		initiodconn(mypid);
		return -1;
	}
	for(i=0; i < initial_entries; i++) {
		if(iodinfo[i].valid != 0) {
			if(iodinfo[i].fd == sockfd) {
				return i;
			}
		}
	}
	return -1;
}

/* Just add the iod information if necessary. Dont try to connect. Return the slot information */
int instantiate_iod_entry(struct sockaddr *saddr_p)
{
	int slot, free_slot = initial_entries;
	if(!iodinfo) {
		if(init_iodtable(0,0) < 0) {
			return -ENOMEM;
		}
	}
	/* Try to search for the host's slot in our table */
repeat:
	/* Did not find it in the table */
	if((slot = search_iodtable(saddr_p, &free_slot)) < 0) {
		if(free_slot < initial_entries) {
			iodinfo[free_slot].valid = 1;
			iodinfo[free_slot].addr = ((struct sockaddr_in *) saddr_p)->sin_addr.s_addr;
			iodinfo[free_slot].port = ((struct sockaddr_in *) saddr_p)->sin_port;
			iodinfo[free_slot].fd = -1;
			iodcnt++;
		}
		/* No free space in the table */
		else {
			if(grow_iodtable() < 0) {
				return -ENOMEM;
			}
			goto repeat;
		}
		slot = free_slot;
	}
	return slot;
}

/* Either add iod information or retrieve the sock fd already stored */
int add_iodtable(struct sockaddr *saddr_p)
{
	int free_slot = initial_entries, slot;
	if(!iodinfo) {
		if(init_iodtable(0,0) < 0) {
			return -ENOMEM;
		}
	}
	/* Try to search for the host's slot in our table */
repeat:
	/* Did not find it in the table */
	if((slot = search_iodtable(saddr_p, &free_slot)) < 0) {
		if(free_slot < initial_entries) {
			iodinfo[free_slot].valid = 1;
			iodinfo[free_slot].addr = ((struct sockaddr_in *) saddr_p)->sin_addr.s_addr;
			iodinfo[free_slot].port = ((struct sockaddr_in *) saddr_p)->sin_port;
			iodinfo[free_slot].fd = -1;
			iodcnt++;
		}
		/* No free space in the table */
		else {
			if(grow_iodtable() < 0) {
				return -ENOMEM;
			}
			goto repeat;
		}
		slot = free_slot;
	}
	/* connect if necessary */
	if(iodinfo[slot].fd < 0) {
		int fd, err;
		struct ireq req;
		struct iack ack;
		fd = iodinfo[slot].fd = new_sock();
		if(fd < 0) {
			return fd;
		}
		if ((err = connect_timeout(fd, saddr_p, sizeof(struct sockaddr),IODCOMM_CONNECT_TIMEOUT_SECS)) < 0) {
			Close(fd);
			iodinfo[slot].fd = -1;
			return err;
		}
		LOG3("Connected to %d:%d on %d\n",((struct sockaddr_in *)saddr_p)->sin_addr.s_addr, ((struct sockaddr_in *)saddr_p)->sin_port, fd);
		req.majik_nr   = IOD_MAJIK_NR;
		req.release_nr = PVFS_RELEASE_NR;
		req.type       = IOD_NOOP;
		req.dsize      = 0;
		err = bsend(fd, &req, sizeof(req));
		if (err < 0) {
			PERROR("unable to send request to %d:%d on %d\n", ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr, ((struct sockaddr_in *)saddr_p)->sin_port, fd);
			Close(fd);
			iodinfo[slot].fd = -1;
			return -1;
		}
		err = brecv_timeout(fd, &ack, sizeof(ack), IODCOMM_BRECV_TIMEOUT_SECS);
		if(err < 0) {
			PERROR("unable to receive ack from %d:%d on %d within timeout\n", ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr, ((struct sockaddr_in *)saddr_p)->sin_port, fd);
			Close(fd);
			iodinfo[slot].fd = -1;
			return -1;
		}
		if (ack.majik_nr != IOD_MAJIK_NR) {
			PERROR("Nonsense ack from %d:%d on %d \n", ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr, ((struct sockaddr_in *)saddr_p)->sin_port, fd);
			Close(fd);
			iodinfo[slot].fd = -1;
			return -1;
		}
		if(ack.release_nr != PVFS_RELEASE_NR) {
			PERROR("release nr from %d:%d on %d is (%d.%d.%d) does not match release nr built into binary (%d.%d.%d)\n", ((struct sockaddr_in *)saddr_p)->sin_addr.s_addr, ((struct sockaddr_in *)saddr_p)->sin_port, fd, ack.release_nr / 10000, (ack.release_nr / 100) % 10, ack.release_nr % 10, PVFS_RELEASE_NR / 10000, (PVFS_RELEASE_NR / 100) % 10, PVFS_RELEASE_NR % 10);
			Close(fd);
			iodinfo[slot].fd = -1;
			return -1;
		}
		/* update fd table */
		if (pfds[fd]) {
			LOG1("fd %d already in use?!?\n", fd);
		}
		pfds[fd] = (fdesc_p)kmalloc(sizeof(fdesc),GFP_KERNEL);
		if(!pfds[fd]) {
			LOG("Could not allocate memory\n");
			Close(fd);
			iodinfo[slot].fd = -1;
			return -ENOMEM;
		}
		pfds[fd]->fs = FS_RESV;
		pfds[fd]->part_p = NULL;
		pfds[fd]->dim_p  = NULL;
		pfds[fd]->fd.ref = -1;
		pfds[fd]->fd.grp_nr = -1;
	}
	return(iodinfo[slot].fd);
}

int badiodfd(int fd)
{
	int i;
	if(!iodinfo) {
		return -EINVAL;
	}
	for (i = 0; i < initial_entries; i++) {
		if (fd == iodinfo[i].fd) {
			iodinfo[i].ref_count = 0; /* anyways since the sock fd is closed, we can reset the ref count */
			Close(fd);
			iodinfo[i].fd = -1;
			/* update fd table */
			if (!pfds[fd] || pfds[fd]->fs != FS_RESV) {
				ERR1("fd %d was not reserved?!?\n", fd);
				return(-1);
			}
			kfree(pfds[fd]);
			pfds[fd] = NULL;
			return 0;
		}
	}
	return -1;
}

/* What a stupid name... */
int shutdown_alliods(void)
{
	int i;
	if(!iodinfo) {
		return -EINVAL;
	}
	for(i=0;i < initial_entries; i++) {
		int fd;
		if(iodinfo[i].valid != 0 && ((fd=iodinfo[i].fd) >= 0)) {
			/* Do a sanity check */
			if(iodinfo[i].ref_count != 0) {
				ERR2("Hmm..slot %d had a non-zero(%d) ref count\n",i, iodinfo[i].ref_count);
			}
			Close(fd);
			/* update fd table */
			if (!pfds[fd] || pfds[fd]->fs != FS_RESV) {
				ERR1("fd %d was not reserved?!?\n", fd);
				return(-1);
			}
			kfree(pfds[fd]);
			pfds[fd] = NULL;
			iodinfo[i].fd = -1;
		}
	}
	return 0;
}

static int initiodconn(pid_t pid)
{
	int i;

	init_iodtable(0, 0);

	for(i=0; i < initial_entries; i++) {
		if(iodinfo[i].fd >= 0) {
			Close(iodinfo[i].fd);
			iodinfo[i].fd = -1;
			iodinfo[i].ref_count = 0;
		}
	}
	iodcnt = 0;
	iodpid = pid;
	return 0;
}

int inc_ref_count(int slot)
{
	if(!iodinfo) {
		return -EINVAL;
	}
	if(slot < 0 || slot >= initial_entries) {
		return -EINVAL;
	}
	/* valid slot */
	if(iodinfo[slot].valid != 0) {
		iodinfo[slot].ref_count++;
		return 0;
	}
	return -EINVAL;
}

int dec_ref_count(int slot)
{
	if(!iodinfo) {
		return -EINVAL;
	}
	if(slot < 0 || slot >= initial_entries) {
		return -EINVAL;
	}
	/* valid slot */
	if(iodinfo[slot].valid != 0) {
		iodinfo[slot].ref_count--;
		return 0;
	}
	return -EINVAL;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 *
 * vim: ts=3
 */
