/*
 * mount.pvfs.c copyright (c) 1997 Clemson University, all rights reserved.
 *
 * Written by Rob Ross, based somewhat on nfs mount code by Rick Sladkey
 * <jrs@world.std.com>.
 *
 * Modified by Dan Nurmi <nurmi@mcs.anl.gov> to get /etc/mtab right.
 * Option parsing code based somewhat on smb mount code.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.eng.clemson.edu
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/mount.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <mntent.h>
#include "pvfs_mount.h"

#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include "sockio.h"
#include "req.h"

#define DEBUG 0
#define dprintf(args...) if (DEBUG) printf(args);

struct pvfs_mount_clargs {
	char *hostname;
	char *dirname;
	char *special;
	char *mountpoint;
	char *amode;
	int port;
};

static int usage(int argc, char **argv);
static int do_mtab(struct mntent *);
static int setup_mntent(struct mntent *, const char *, const char *, 
const char *, const char *, int, int);
static int parse_args(int argc, char *argv[], struct pvfs_mount_clargs *);
static int resolve_addr(char *, struct in_addr *);
static int nb_mgr_ping(char *name, int port, int timeout_secs, int retries);
static void print_retry_msg(char *name);


int main(int argc, char **argv)
{
	int ret, flags = 0;
	char type[] = "pvfs";
	char *host_p;
	struct pvfs_mount data;
	struct pvfs_mount_clargs clargs;
	struct mntent myment;
	struct in_addr haddr;
	struct stat sb, mntpt;
  
	/* gobble up some memory */
	clargs.hostname = malloc(sizeof(char[PVFS_MGRLEN]));
	clargs.dirname = malloc(sizeof(char[PATH_MAX]));
	clargs.special = malloc(sizeof(char[PVFS_MGRLEN + PATH_MAX + 1]));
	clargs.mountpoint = malloc(sizeof(char[PATH_MAX]));
	clargs.amode = malloc(sizeof(char[3]));

	if (argc < 3) {
		fprintf(stderr, "mount.pvfs: too few arguments\n");
		usage(argc, argv);
		return(1);
	}

	if ((ret = parse_args(argc, argv, &clargs))) {
		fprintf(stderr, "ERROR: problem parsing cmdline args\n");
		return(ret);
	}

	/* look up the hostname (could use DNS) */
	if ((ret = resolve_addr(clargs.hostname, &haddr))) {
		return(ret);
	}

	/* convert address back to dotted decimal notation to pass into kernel */
	host_p = inet_ntoa(haddr);
	strncpy(clargs.hostname, host_p, PVFS_MGRLEN);

	dprintf("hostname after transform | hostname=%s\n", clargs.hostname);

	data.info_magic = 0;
	data.flags   = 0;
	data.port    = clargs.port;
	strncpy(data.mgr, clargs.hostname, PVFS_MGRLEN);
	strncpy(data.dir,  clargs.dirname,  PVFS_DIRLEN);

	flags |= 0xC0ED0000;

	/* verify that local mountpoint exists and is a directory */
	if (stat(clargs.mountpoint, &mntpt) < 0 || !S_ISDIR(mntpt.st_mode)) {
		fprintf(stderr, "mount.pvfs: invalid mount point %s (should be a local directory)\n",
				  clargs.mountpoint);
		return 1;
	}

	/* ping the manager before trying to mount */
	if (nb_mgr_ping(data.mgr, data.port, 20, 2) != 0) {
		fprintf(stderr, "mount.pvfs: server %s:%d not responding, giving up\n",
				  data.mgr, data.port);
		return 1;
	}

	dprintf("calling mount(special = %s, mountpoint = %s, type = %s,"
			  " flags = %x, data = %lx)\n", clargs.special, clargs.mountpoint, type, flags, (unsigned long) &data);
	if ((ret = mount(clargs.special, clargs.mountpoint, type, flags, (void *) &data)) < 0)
	{
		perror("mount.pvfs");
		fprintf(stderr, "mount.pvfs: server %s alive, but mount failed (invalid metadata directory name?)\n",
				  data.mgr);
		return (ret);
	}

	if ((ret = setup_mntent(&myment, clargs.special, clargs.mountpoint, type, clargs.amode, 0, 0))) {
		return(ret);
	}

   /* Leave mtab alone if it is a link */
	if (lstat(MTAB, &sb) == 0 && S_ISLNK(sb.st_mode)) return ret;
	
	ret = do_mtab(&myment);

	return(ret);
} /* end of main() */


/* FUNCTIONS */
int new_sock() 
{
	static int p_num = -1; /* set to tcp protocol # on first call */
	struct protoent *pep;

   if (p_num == -1) {
#ifdef HAVE_SCYLD
		if ((pep = getprotobyname("tcp")) == NULL) {
			if(errno == ENOENT){
				/* On Scyld systems we use a in.h define for TCP when the
				 * /etc/protocols file is not present locally.
				 */
				p_num = IPPROTO_TCP;
			}
			else{
				perror("Kernel does not support tcp.");
				return(-1);
			}
		}
		else{
			p_num = pep->p_proto;
		}
#else
		if ((pep = getprotobyname("tcp")) == NULL) {
			perror("Kernel does not support tcp");
			return(-1);
		}
		p_num = pep->p_proto;
#endif
	}
	return(socket(AF_INET, SOCK_STREAM, p_num));
}


int init_sock(struct sockaddr *saddrp, char *name, int service)
{
	struct hostent *hep;

	bzero((char *)saddrp,sizeof(struct sockaddr_in));
	if (name == NULL) {
		if ((hep = gethostbyname("localhost")) == NULL) {
			return(-1);
		}
	}
	else if ((hep = gethostbyname(name)) == NULL) {
		return(-1);
	}
	((struct sockaddr_in *) saddrp)->sin_family = AF_INET;
	((struct sockaddr_in *)saddrp)->sin_port = htons((u_short)service);
	bcopy(hep->h_addr, (char *)&(((struct sockaddr_in *)saddrp)->sin_addr),
		hep->h_length);
	return(0);
}

/* connect_timeout()
 *
 * Attempts to do nonblocking connect() until a timeout occurs.  Useful for
 * recovering from deadlocks...
 *
 * I apologize for all the goto's, but I hate to every syscall in a while()
 * just for the EINTR case!  
 */
int connect_timeout(int s, struct sockaddr *saddrp, int len, int time_secs)
{
	struct pollfd pfds;
	int ret, oldfl, err, val, val_len;

	/* set our socket to nonblocking */
	oldfl = fcntl(s, F_GETFL, 0);
	if (!(oldfl & O_NONBLOCK)) fcntl(s, F_SETFL, oldfl | O_NONBLOCK);

connect_timeout_connect_restart:
	ret = connect(s, saddrp, len);
	if (ret < 0) {
		if (errno == EINTR) goto connect_timeout_connect_restart;
		if (errno != EINPROGRESS) goto connect_timeout_err;

		pfds.fd = s;
		pfds.events = POLLOUT;
connect_timeout_poll_restart:
		ret = poll(&pfds, 1, time_secs * 1000);
		if (ret < 0) {
			if (errno == EINTR) goto connect_timeout_poll_restart;
			else goto connect_timeout_err;
		}
		if (ret == 0) {
			/* timed out */
			err = errno;
			close(s); /* don't want it to keep trying */
			errno = err;
			goto connect_timeout_err;
		}
	}

	/* _apparent_ success -- check if connect really completed */
	val_len = sizeof(val);
	ret = getsockopt(s, SOL_SOCKET, SO_ERROR, &val, &len);
	if (val != 0) {
		errno = val;
		goto connect_timeout_err;
	}

	/* success -- make socket blocking again and return */
	fcntl(s, F_SETFL, oldfl & (~O_NONBLOCK));
	return 0;

connect_timeout_err:
	/* save errno, set flags on socket back to what they were, return -1 */
	err = errno;
	fcntl(s, F_SETFL, oldfl);
	errno = err;
	return -1;
}

/* nb_mgr_ping() 
 *
 * sends a noop request to a manager hoping to elicit a
 * response.  this is done in a non-blocking manner.
 *
 * not sure i'm handling retries exactly right (wrt number of retries),
 * but it definitely gets decremented...
 *
 * Returns 0 on success, 1 on failure.
 */
static int nb_mgr_ping(char *name, int port, int timeout_secs, int retries)
{
	int s, oldfl, ret, index, size;
	struct sockaddr saddr;
	fd_set set;
	struct timeval timeout;
	struct mreq req;
	struct mack ack;
	char *buf;

	if (retries < 0) retries = 0;

	/* fill in sockaddr structure (sockio pvfs call) */
	if (init_sock(&saddr, name, port) < 0) return -1;

	s = new_sock();

	while (retries >= 0) {
		if (s < 0) return 1;
		ret = connect_timeout(s, &saddr, sizeof(saddr), timeout_secs);
		if (ret == 0) break;
		
		close(s);
		s = new_sock();
		retries--;
	}
	if (retries < 0) {
		close(s);
		return 1;
	}

	/* set our socket to nonblocking */
	oldfl = fcntl(s, F_GETFL, 0);
	if (!(oldfl & O_NONBLOCK)) fcntl(s, F_SETFL, oldfl | O_NONBLOCK);

	/* prepare request, send it */
	req.majik_nr   = MGR_MAJIK_NR;
	req.release_nr = PVFS_RELEASE_NR;
	req.type       = MGR_NOOP;
	req.dsize      = 0;

	buf = (char *) &req;
	size = sizeof(struct mreq);
	ret = write(s, buf, size);

	/* in normal case we will have sent the whole thing and be done */
	if (ret < size) {
		index = 0;
		/* not normal case :( */
		while (retries >= 0) {
			print_retry_msg(name);
			index += ret > 0 ? ret : 0; /* only add if positive */
			if (index == size) break; /* done with write */	
			retries--;

			if (ret < 0 && errno != EAGAIN) {
				perror("write");
				goto nb_mgr_ping_err;
			}

			/* select again */
			FD_ZERO(&set);
			FD_SET(s, &set);
			timeout.tv_usec = 0;
			timeout.tv_sec = timeout_secs;
			ret = select(s+1, NULL, &set, NULL, &timeout);
			if (ret == 0) continue; /* not ready yet */
			if (ret < 0) {
				perror("select (writing)");
				goto nb_mgr_ping_err;
		  	}
			
			ret = write(s, buf + index, size - index);	
		}
	}
	if (retries < 0) goto nb_mgr_ping_err;

	/* read response */
	index = 0;
	size = sizeof(struct mack);

	while (retries >= 0) {
		FD_ZERO(&set);
		FD_SET(s, &set);
		timeout.tv_usec = 0;
		timeout.tv_sec = timeout_secs;
		ret = select(s+1, &set, NULL, NULL, &timeout);
		if (ret == 0) {
			print_retry_msg(name);
			retries--;
			continue;
		}
		if (ret < 0) {
			perror("select (reading)");
			goto nb_mgr_ping_err;
		}

		ret = read(s, (char *)(&ack) + index, size);
		if (ret < 0 && errno != EAGAIN) {
			perror("read");
			goto nb_mgr_ping_err;
		}
		index += ret > 0 ? ret : 0;
		if (index == size) break;
		print_retry_msg(name);
		retries--;
	}
	if (retries < 0) goto nb_mgr_ping_err;

	/* whew!  made it through all that mess; now let's see if the response
	 * had the right magic value
	 */
	if (ack.majik_nr != MGR_MAJIK_NR) {
		goto nb_mgr_ping_err;
	}
	if (ack.release_nr != PVFS_RELEASE_NR) {
		fprintf(stderr, "mount.pvfs: release number from mgr at %s:%d (%d.%d.%d) does not match release number built into this binary (%d.%d.%d).\n",
				  name,
				  port,
				  ack.release_nr / 10000,
				  (ack.release_nr / 100) % 10,
				  ack.release_nr % 10,
				  PVFS_RELEASE_NR / 10000,
				  (PVFS_RELEASE_NR / 100) % 10,
				  PVFS_RELEASE_NR % 10);
		fprintf(stderr, "mount.pvfs: an administrator needs to verify that all PVFS servers and clients are at the same release number.\n");
		fflush(stderr);
		goto nb_mgr_ping_err;
	}

	/* is closing a nonblocking socket an issue? */
	close(s);
	return 0;

 nb_mgr_ping_err:
	close(s);
	return 1;
}

static void print_retry_msg(char *name)
{
	fprintf(stderr, "mount.pvfs: server %s not responding, still trying\n", 
			  name);
	fflush(stderr);
	return;
}


/* resolve_addr() 
 *
 * returns struct in_addr holding address of a host specified by the character
 * string name, which can either be a host name or IP address in dot notation.
 */
static int resolve_addr(char *name, struct in_addr *addr)
{
   struct hostent *hep;

   if (!inet_aton(name, addr)) {
		if (!(hep = gethostbyname(name))) {
			fprintf(stderr, "ERROR: cannot resolve remote hostname\n");
			return(1);
		} else {
			bcopy(hep->h_addr, (char *)addr, hep->h_length);
			return(0);
		}
   }
   return(0);
} /* end of resolve_addr() */

/* 
 * parse_args() 
 *
 * function takes argc/argv pair and a pointer to a struct pvfs_mount_clargs.
 * The struct contains variables that need to be filled with information from
 * the commandline to be later used back in main().  
 */
static int parse_args(int argc, char *argv[], 
struct pvfs_mount_clargs *in_clargs) 
{
	int opt, count;
	char *index,
		*subopt,
		*subval,
		*mopts = malloc(sizeof(char[255])),
		*hostdir = malloc(sizeof(char[strlen(argv[1])+1]));

	/* separate hostname and dirname from 'hostname:dirname' format */
	strcpy(hostdir, argv[1]);
	if ((index = (strchr(hostdir, ':')))) {
		*index = '\0';

		strcpy(in_clargs->hostname, hostdir);

		index++;
		strcpy(in_clargs->dirname, index);

	}  else /* not good */ {
		fprintf(stderr, 
				  "mount.pvfs: directory to mount not in host:dir format\n");
		free(mopts);
		free(hostdir);
		return(1);
	}

	strcpy(in_clargs->special, argv[1]);
	strcpy(in_clargs->mountpoint, argv[2]);

	argc -= 2;
	argv += 2;

	/* Set default values */
	in_clargs->port = 3000;
	strcpy(in_clargs->amode, "rw");
  
	/* Start parsage */
	while ((opt = getopt (argc, argv, "o:")) != EOF)
	{
		switch (opt)
		{
		case 'o':
			strcpy(mopts, optarg);
			count = 0;

			index = (char *)strtok(mopts, ",");
			while(index != NULL) {
				if ((subopt = (char*)strchr(index, '=')) != NULL) {
					subval = subopt + 1;
					*subopt = '\0';
	      
					/* 
					 * Start ifelse that sets internal variables 
					 * equal to commandline arguments of the form 
					 * 'key=val' 
					 */
					if (!strcmp(index, "port")) {
						in_clargs->port = atoi(subval);
					}


				} else {
					/* 
					 * Start ifelse that sets internal variables 
					 * equal to commandline arguments of the form 
					 * 'val' 
					 */
					if (!strcmp(index, "rw")) {
						strcpy(in_clargs->amode, "rw");
					} 
					else if (!strcmp(index, "ro")) {
						strcpy(in_clargs->amode, "ro");
					}
				}
				index = (char*)strtok(NULL, ",");
			}
			break;
		default:
			free(mopts);
			free(hostdir);	  
			return 1;
		}
	}

	free(mopts);
	free(hostdir);

	dprintf("Parsed Args: hostname=%s, remote_dirname=%s, local_mountpoint=%s, special=%s, amode=%s, port=%d\n", in_clargs->hostname, in_clargs->dirname, in_clargs->mountpoint, in_clargs->special, in_clargs->amode, in_clargs->port);
	return 0;
}

static int usage(int argc, char **argv)
{
	fprintf(stderr,
			  "mount.pvfs: mgr_host:metadata_dir directory [-o <options>]\n            Valid options are port=PORT, rw. Multiple options\n            may be comma delimited.\n");
	return 0;
} /* end of USAGE() */

/* do_mtab - Given a pointer to a filled struct mntent,
 * add an entry to /etc/mtab.
 *
 */
static int do_mtab(struct mntent *myment) 
{
	struct mntent *ment;
	FILE *mtab;
	FILE *tmp_mtab;

	mtab = setmntent(MTAB, "r");
	tmp_mtab = setmntent(TMP_MTAB, "w");

	if (mtab == NULL) {
		fprintf(stderr, "ERROR: couldn't open "MTAB" for read\n");
		endmntent(mtab);
		endmntent(tmp_mtab);
		return 1;
	} else if (tmp_mtab == NULL) {
		fprintf(stderr, "ERROR: couldn't open "TMP_MTAB" for write\n");
		endmntent(mtab);
		endmntent(tmp_mtab);
		return 1;
	}

	while((ment = getmntent(mtab)) != NULL) {
		if (strcmp(myment->mnt_dir, ment->mnt_dir) != 0) {
			if (addmntent(tmp_mtab, ment) == 1) {
				fprintf(stderr, "ERROR: couldn't add entry to"TMP_MTAB"\n");
				endmntent(mtab);
				endmntent(tmp_mtab);
				return 1;
			}
		}
	}

	endmntent(mtab);

	if (addmntent(tmp_mtab, myment) == 1) {
		fprintf(stderr, "ERROR: couldn't add entry to "TMP_MTAB"\n");
		return 1;
	}

	endmntent(tmp_mtab);

	if (rename(TMP_MTAB, MTAB)) {
		fprintf(stderr, "ERROR: couldn't rename "TMP_MTAB" to "MTAB"\n");
		return 1;
	}

	if (chmod(MTAB, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) < 0) {
		if (errno != EROFS) {
			int errsv = errno;
			fprintf(stderr, "mount: error changing mode of %s: %s",
					  MTAB, strerror (errsv));
		}
		return(1);
	}
	return 0;
} /* end of do_mtab */


/* setup_mntent - allocate space for members of mntent, fill in the data
 * structure
 */
static int setup_mntent(struct mntent *myment, const char *fsname, 
const char *dir, const char *type, const char *opts, int freq, int passno)
{
	myment->mnt_fsname = malloc(sizeof(char[strlen(fsname)+1]));
	myment->mnt_dir = malloc(sizeof(char[strlen(dir)+1]));
	myment->mnt_type = malloc(sizeof(char[strlen(type)+1]));
	myment->mnt_opts = malloc(sizeof(char[strlen(opts)+1]));

	if ((myment->mnt_fsname == NULL) || ( myment->mnt_dir == NULL)
		 || (myment->mnt_type == NULL) || (myment->mnt_opts == NULL))
	{
		fprintf(stderr, "ERROR: cannot allocate memory\n");
		return 1;
	}

	strcpy(myment->mnt_fsname, fsname);
	strcpy(myment->mnt_dir, dir);
	strcpy(myment->mnt_type, type);
	strcpy(myment->mnt_opts, opts);
	myment->mnt_freq = freq;
	myment->mnt_passno = passno;
  
	return 0;
} /* end setup_mntent */

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 *
 * vim: ts=3
 */
