#ifndef __PVFS_KERNEL_CONFIG_H
#define __PVFS_KERNEL_CONFIG_H

/*
 * pvfs_kernel_config.h copyright (c) 1999 Rob Ross and Phil Carns.
 * all rights reserved.
 *
 * Written by Rob Ross and Phil Carns.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

#include "config.h"

/* Kernel configuration issues */
#ifdef __KERNEL__

/* Some of these conditional statements were borrowed from Don Becker's
 * 3c509.c module code.
 */
#include <linux/config.h>
#include <linux/version.h>

/* define __SMP__ if we detect SMP kernel running at configure time */
#if defined(HAVE_SMP_KERNEL) && ! defined(__SMP__)
#define __SMP__  
#endif 
#if !defined(HAVE_SMP_KERNEL) && defined(__SMP__)
#undef __SMP__
#endif

/* define MODVERSIONS if we detect running kernel uses them at 
 * configure time
 */
#if defined(HAVE_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#if !defined(HAVE_MODVERSIONS) && defined(MODVERSIONS)
#undef MODVERSIONS
#endif

#if defined(MODVERSIONS) && defined(HAVE_LINUX_MODVERSIONS_H)
#include <linux/modversions.h>
#endif

/* This include moved down to ensure that it follows the inclusion of
 * modversions.h if it is used.  Otherwise we run into symbol problems
 * on new 2.4 kernels
 */
#include <linux/module.h>

#endif /* __KERNEL__ */

/* Decide if we are going to be able to support mapping user buffers
 * into kernel space or not.  This should always work on 2.4 kernels
 * unless HIGHMEM is enabled */
#if !defined(HAVE_HIGHMEM)
#define HAVE_PVFS_BUFMAP 1
#else
#undef HAVE_PVFS_BUFMAP
#endif

/* DEBUGGING VALUES */
enum {
	D_SUPER = 1,  /* superblock calls */
	D_INODE = 2,  /* inode calls */
	D_FILE = 4,  /* file calls */
	D_DIR = 8,  /* directory calls */
	D_LLEV = 16, /* low level calls */
	D_LIB = 32,
	D_UPCALL = 64,
	D_PSDEV = 128,
	D_PIOCTL = 256,
	D_SPECIAL = 512,
	D_TIMING = 1024,
	D_DOWNCALL = 2048,
	D_ENTRY = 4096,
	D_PVFSD = 8192,
	D_MOD = 16384,
	D_ALLOC = 32768
};
 
#ifdef __KERNEL__
#define PDEBUG(mask, format...)                                \
  do {                                                            \
    if (pvfs_debug & mask) {                                      \
      printk("(%s, %d): ",  __FILE__, __LINE__);                  \
      printk(format);                                       \
	 }                                                             \
  } while (0) ;
#define PERROR(format...)                                      \
  do {                                                            \
    printk("(%s, %d): ",  __FILE__, __LINE__);                    \
    printk(format);                                         \
  } while (0) ;
#define PENTRY                                                    \
  if(pvfs_debug & D_ENTRY)                                        \
	 printk("Process %d entered %s\n",current->pid,__FUNCTION__)
#define PEXIT                                                     \
  if(pvfs_debug & D_ENTRY)                                        \
    printk("Process %d leaving %s\n",current->pid,__FUNCTION__) 

#else
#define PDEBUG(mask, format...)                                \
  do {                                                            \
    if (pvfs_debug & mask) {                                      \
      printf("(%s, %d): ",  __FILE__, __LINE__);                  \
      printf(format);                                       \
    }                                                             \
  } while (0) ;
#define PERROR(format...)                                      \
  do {                                                            \
    printf("(%s, %d): ",  __FILE__, __LINE__);                    \
    printf(format);                                         \
	 fflush(stdout);                                               \
  } while (0) ;
#endif

/* Buffer allocation values:
 * PVFS_BUFFER_STATIC - allocates a static buffer and holds it for the
 *   duration the module is loaded
 * PVFS_BUFFER_DYNAMIC - allocates a memory region in the kernel for the
 *   duration of each request
 * PVFS_BUFFER_MAPPED - maps the user's buffer into kernel space,
 *   eliminating additional memory use and a copy
 */
enum {
	PVFS_BUFFER_STATIC = 1,
	PVFS_BUFFER_DYNAMIC = 2,
	PVFS_BUFFER_MAPPED = 3
};

/* Values:
 * PVFS_DEFAULT_IO_SIZE - default buffer size to service at one time.  This
 *   is used to prevent very large transfers from grabbing all available
 *   physical memory.
 * PVFS_DEFAULT_BUFFER - default buffering technique in the kernel.
 * PVFS_OPT_IO_SIZE - value returned on statfs; determines the transfer
 *   size used by many common utilities?  Well, maybe not...we still
 *   seem to be getting 64K transfers.
 * PVFS_DEFAULT_DEBUG_MASK - used in pvfs_mod.c and pvfsd.c as the
 *   default mask for debugging.  This can be changed for the module with
 *   the "debug" parameter on insmod.
 * PVFSD_NOFILE - not currently used; will eventually be the number of
 *   file descriptors the pvfsd allows itself to use
 * PVFSD_MAJOR - major number for pvfsd device file
 */
enum {
	PVFS_DEFAULT_IO_SIZE = 16*1024*1024,
	PVFS_MIN_IO_SIZE = 64*1024,
	PVFS_DEFAULT_BUFFER = PVFS_BUFFER_DYNAMIC,
	PVFS_OPT_IO_SIZE = 128*1024, 
	PVFS_DEFAULT_DEBUG_MASK = 0,
	PVFSD_NOFILE = 65536,
	PVFSD_MAJOR = 60 /* experimental number */
};

#endif
/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
